﻿using System;
using System.Linq;
using Microsoft.Scripting;
using Microsoft.Scripting.Hosting;
using Microsoft.Scripting.Runtime;

namespace Hardcodet.Scripting
{
    /// <summary>
    /// A symbol dictionary that provides an fixed set of
    /// symbols through a <see cref="BaseScope"/>. As new variables
    /// are not added to the <see cref="BaseScope"/>, these cached
    /// symbols can be reused across different scopes.
    /// </summary>
    public class SymbolCache : CustomSymbolDictionary
    {
      /// <summary>
      /// A script scope that provides a reusable set of symbols.
      /// Any variables that are being created by the <see cref="ScriptScope"/>
      /// that owns this cache are not stored within <see cref="BaseScope"/>,
      /// but the parent scope's own symbol dictionary.
      /// </summary>
      public ScriptScope BaseScope { get; private set; }


      /// <summary>
      /// Creates a new cache instance
      /// </summary>
      /// <param name="baseScope">A reusable <see cref="ScriptScope"/> that provides
      /// a set of symbols that are supposed to be used across several scopes.</param>
      /// <exception cref="ArgumentNullException">If <paramref name="baseScope"/>
      /// is a null reference.</exception>
      public SymbolCache(ScriptScope baseScope)
      {
        if (baseScope == null) throw new ArgumentNullException("baseScope");
        BaseScope = baseScope;
      }



      /// <summary>
      /// Invoked if a given variable or symbol is being requested. This method
      /// tries to find the requested item in the underlying <see cref="BaseScope"/>.
      /// </summary>
      /// <param name="key"></param>
      /// <param name="value"></param>
      /// <returns>True if the <see cref="BaseScope"/> provides the requested
      /// symbol.</returns>
      protected override bool TryGetExtraValue(SymbolId key, out object value)
      {
        //return the key from the base scope, if possible.
        lock (BaseScope)
        {
          return BaseScope.TryGetVariable(SymbolTable.IdToString(key), out value);
        }
      }



      /// <summary>
      /// Gets a list of the extra keys that are cached by the the optimized
      /// implementation of the module.
      /// </summary>
      public override SymbolId[] GetExtraKeys()
      {
        lock (BaseScope)
        {
          return BaseScope.GetItems().Select(pair =>
            SymbolTable.StringToId(pair.Key)).ToArray();
        }
      }


      /// <summary>
      /// Tries to set the extra value and return true if the specified key
      /// was found in the list of extra values.<br/>
      /// Any attempts to store extra values are being denied, which causes
      /// them to be stored in the scope itself rather than the local
      /// <see cref="BaseScope"/>. This ensures that runtime variables are
      /// not shared between different instances of the cache.
      /// </summary>
      /// <param name="key">The key that is used to store the submitted
      /// value.</param>
      /// <param name="value">Value to be cached.</param>
      /// <returns>Always false because runtime symbols are not supposed
      /// to be stored within the cache. This causes the value to be stored
      /// within the internal dictionary of the base class.</returns>
      protected override bool TrySetExtraValue(SymbolId key, object value)
      {
        return false;
      }

    }
}